#![feature(c_size_t)]

use std::fs::File;
use std::io::{Cursor, Read, Seek, SeekFrom, Write};
use std::path::PathBuf;

use clap::{command, value_parser, Arg, ArgAction, ArgMatches, Command};
use color_eyre::eyre::{self, Context};
use color_eyre::Result;
use tracing_error::ErrorLayer;
use tracing_subscriber::prelude::*;
use tracing_subscriber::EnvFilter;

mod binary;
mod oodle;
mod types;

use binary::*;
use oodle::{Oodle, CHUNK_SIZE};
use types::*;

const LIB_NAME: &str = "oo2core_8_win64";

#[tracing::instrument]
fn main() -> Result<()> {
    color_eyre::install()?;

    let matches = command!()
        .about("Perform Oodle 2 compression and decompression for Darktide Bitsquid bundles.")
        .subcommand_required(true)
        .arg(
            Arg::new("verbosity")
                .help("Set verbosity for the Oodle library.")
                .global(true)
                .short('v')
                .long("verbose")
                .default_value("none")
                .value_parser(value_parser!(OodleLZ_Verbosity)),
        )
        .arg(
            Arg::new("library")
                .help(
                    "Library to load. This may either be:\n\
                        - A library name that will be searched for in the system's default paths.\n\
                        - A file path relative to the current working directory.\n\
                        - An absolute file path.",
                )
                .long("library")
                .default_value(LIB_NAME),
        )
        .subcommand(
            Command::new("compress")
                .about("Compress an arbitrary binary stream in chunks.")
                .arg(
                    Arg::new("offset")
                        .help(
                            "The offset of the compressed data stream within the full bundle file.\
                            \nThis needs to be known during decompression to correctly calculate \
                            the padding within each chunk.",
                        )
                        .long("offset")
                        .required(true)
                        .value_parser(value_parser!(u64)),
                )
                .arg(
                    Arg::new("in_file")
                        .help("The input file")
                        .required(true)
                        .value_parser(value_parser!(PathBuf)),
                )
                .arg(
                    Arg::new("out_file")
                        .help("The output file")
                        .required(true)
                        .value_parser(value_parser!(PathBuf)),
                ),
        )
        .subcommand(
            Command::new("decompress")
                .about("Decompress a stream of chunks.")
                .arg(
                    Arg::new("offset")
                        .help(
                            "The offset of the compressed data stream within the full bundle file.\
                            \nThis needs to be known during decompression to correctly calculate \
                            the padding within each chunk.",
                        )
                        .long("offset")
                        .required(true)
                        .value_parser(value_parser!(u64)),
                )
                .arg(
                    Arg::new("fuzz_safe")
                        .long("fuzz-safe")
                        .action(ArgAction::SetTrue),
                )
                .arg(
                    Arg::new("check_crc")
                        .long("check-crc")
                        .action(ArgAction::SetTrue),
                )
                .arg(
                    Arg::new("in_file")
                        .help("The input file")
                        .required(true)
                        .value_parser(value_parser!(PathBuf)),
                )
                .arg(
                    Arg::new("out_file")
                        .help("The output file")
                        .required(true)
                        .value_parser(value_parser!(PathBuf)),
                ),
        )
        .get_matches();

    {
        let fmt_layer = tracing_subscriber::fmt::layer().pretty();
        let filter_layer =
            EnvFilter::try_from_default_env().or_else(|_| EnvFilter::try_new("info"))?;

        tracing_subscriber::registry()
            .with(filter_layer)
            .with(fmt_layer)
            .with(ErrorLayer::new(
                tracing_subscriber::fmt::format::Pretty::default(),
            ))
            .init();
    }

    let lib = {
        let name = matches.get_one::<String>("library").unwrap();
        let lib = Oodle::new(name).wrap_err_with(|| format!("Failed to load library {}", name))?;
        tracing::info!("Using library {}", name);
        lib
    };

    match matches.subcommand() {
        Some(("compress", sub_matches)) => compress(lib, sub_matches),
        Some(("decompress", sub_matches)) => decompress(lib, sub_matches),
        _ => unreachable!(
            "clap is configured to require a subcommand, and they're all handled above"
        ),
    }
}

fn skip_padding<S>(stream: &mut S, offset: u64) -> Result<()>
where
    S: Seek,
{
    let pos = stream.stream_position()? + offset;
    let padding_size = 16 - (pos % 16);

    tracing::trace!(pos, padding_size);

    if padding_size < 16 && padding_size > 0 {
        stream.seek(SeekFrom::Current(padding_size as i64))?;
    }

    Ok(())
}

#[tracing::instrument(skip_all)]
fn decompress(lib: Oodle, matches: &ArgMatches) -> Result<()> {
    let in_path = matches.get_one::<PathBuf>("in_file").unwrap();
    let out_path = matches.get_one::<PathBuf>("out_file").unwrap();

    tracing::debug!("Reading from '{}'", in_path.display());
    tracing::debug!("Writing to '{}'", out_path.display());

    let fuzz_safe = OodleLZ_FuzzSafe::from(matches.get_flag("fuzz_safe"));
    let check_crc = OodleLZ_CheckCRC::from(matches.get_flag("check_crc"));
    let verbosity = *matches
        .get_one::<OodleLZ_Verbosity>("verbosity")
        .unwrap_or(&OodleLZ_Verbosity::None);

    tracing::trace!(?fuzz_safe, ?check_crc, ?verbosity);

    let offset = *matches.get_one::<u64>("offset").unwrap();

    tracing::debug!(offset);

    let mut in_file = File::open(in_path)
        .wrap_err_with(|| format!("failed to open input file: {}", in_path.display()))?;
    let mut out_file = File::create(out_path)
        .wrap_err_with(|| format!("failed to open output file: {}", out_path.display()))?;

    let num_chunks = read_u32(&mut in_file)?;
    skip_padding(&mut in_file, offset)?;

    tracing::debug!(num_chunks);

    for chunk_index in 0..num_chunks {
        let _span = tracing::debug_span!("Decompressing chunk", chunk_index);
        let chunk_size = read_u32(&mut in_file)? as usize;

        tracing::trace!(chunk_size);

        skip_padding(&mut in_file, offset)?;

        let mut compressed_buffer = vec![0u8; CHUNK_SIZE];
        read_up_to(&mut in_file, &mut compressed_buffer)?;

        // TODO: Optimize to not reallocate?
        let raw_buffer = lib.decompress(&compressed_buffer, fuzz_safe, check_crc, verbosity)?;

        out_file.write_all(&raw_buffer)?;
    }

    Ok(())
}

#[tracing::instrument(skip_all)]
fn compress(lib: Oodle, matches: &ArgMatches) -> Result<()> {
    let in_path = matches.get_one::<PathBuf>("in_file").unwrap();
    let out_path = matches.get_one::<PathBuf>("out_file").unwrap();

    tracing::info!("Reading from '{}'", in_path.display());
    tracing::info!("Writing to '{}'", out_path.display());

    let offset = *matches.get_one::<u64>("offset").unwrap();

    tracing::debug!(offset);

    let mut out_file = File::create(out_path)
        .wrap_err_with(|| format!("failed to open output file: {}", out_path.display()))?;

    let mut r = {
        let mut in_file = File::open(in_path)
            .wrap_err_with(|| format!("failed to open input file: {}", in_path.display()))?;

        // TODO: Is it faster to check metadata for size and pre-allocate?
        let mut buf = Vec::new();
        in_file
            .read_to_end(&mut buf)
            .wrap_err("failed to read input file into memory")?;
        Cursor::new(buf)
    };

    // I'm not sure how Fatshark does this efficiently.
    // Within each chunk, there is a padding whos size depends on the position where the padding is
    // written.
    // And that position is affected by both the size of previous chunks and the total number of
    // chunks, since list of chunk sizes is written before all chunks.
    // Because of that, all chunks have to be compressed and buffered before any of them can be
    // written to the output.
    let mut chunks: Vec<Vec<u8>> = Vec::new();

    let mut chunk_index = 0;

    loop {
        let _span = tracing::info_span!("decompress_chunk", chunk_index);

        let mut raw_buffer = vec![0u8; CHUNK_SIZE];
        let read = read_up_to(&mut r, &mut raw_buffer)?;

        let is_eof = read < CHUNK_SIZE;

        tracing::debug!(is_eof);

        {
            let remaining = CHUNK_SIZE - read;
            if remaining >= 4 {
                for val in raw_buffer.iter_mut().skip(remaining).take(4) {
                    *val = 0x66;
                }
            } else {
                eyre::bail!(
                    "Not enough space left in buffer to add end marker. \
                    Don't know how to proceed."
                );
            }
        }

        let compressed = lib
            .compress(raw_buffer)
            .wrap_err_with(|| format!("failed to compress chunk {}", chunk_index))?;

        chunks.push(compressed);

        if is_eof {
            break;
        }

        chunk_index += 1;
    }

    let out = &mut out_file;

    write_u32(out, chunks.len() as u32)?;

    {
        let _span = tracing::trace_span!("write_chunk_sizes");
        for (chunk_index, chunk) in chunks.iter().enumerate() {
            tracing::trace!(chunk_index);
            write_u32(out, chunk.len() as u32)?;
        }
    }
    write_padding(out, offset)?;

    let unpacked_size = chunks.iter().fold(0, |total, chunk| total + chunk.len());
    write_u32(out, unpacked_size as u32)?;
    write_u32(out, 0)?;

    {
        let _span = tracing::trace_span!("write_chunk_data");
        for (chunk_index, chunk) in chunks.iter().enumerate() {
            tracing::trace!(chunk_index);
            write_u32(out, chunk.len() as u32)?;
            write_padding(out, offset)?;
            out.write_all(chunk)
                .wrap_err("failed to write chunk data")?;
        }
    }

    Ok(())
}
