use std::ffi::OsStr;
use std::ptr;

use color_eyre::eyre;
use color_eyre::Result;
use libloading::{Library, Symbol};

use crate::types::*;

// Hardcoded chunk size of Bitsquid's bundle compression
pub const CHUNK_SIZE: usize = 512 * 1024;

pub struct Oodle {
    lib: Library,
}

impl Oodle {
    pub fn new<P>(lib: P) -> Result<Self>
    where
        P: AsRef<OsStr>,
    {
        let lib = unsafe { Library::new(lib)? };
        // TODO: Register librarie's printf
        Ok(Self { lib })
    }

    #[tracing::instrument(name = "Oodle::decompress", skip(self, data))]
    pub fn decompress<I>(
        &self,
        data: I,
        fuzz_safe: OodleLZ_FuzzSafe,
        check_crc: OodleLZ_CheckCRC,
        verbosity: OodleLZ_Verbosity,
    ) -> Result<Vec<u8>>
    where
        I: AsRef<[u8]>,
    {
        let data = data.as_ref();
        let mut out = vec![0; CHUNK_SIZE];

        let ret = unsafe {
            let decompress: Symbol<OodleLZ_Decompress> = self.lib.get(b"OodleLZ_Decompress")?;

            decompress(
                data.as_ptr() as *const _,
                data.len(),
                out.as_mut_ptr() as *mut _,
                out.len(),
                fuzz_safe,
                check_crc,
                verbosity,
                ptr::null_mut(),
                0,
                ptr::null_mut(),
                ptr::null_mut(),
                ptr::null_mut(),
                0,
                OodleLZ_Decode_ThreadPhase::UNTHREADED,
            )
        };

        if ret < CHUNK_SIZE as u64 {
            eyre::bail!("Failed to decompress chunk.");
        }

        Ok(out)
    }

    #[tracing::instrument(name = "Oodle::compress", skip(self, data))]
    pub fn compress<I>(&self, data: I) -> Result<Vec<u8>>
    where
        I: AsRef<[u8]>,
    {
        let raw = data.as_ref();
        // TODO: Query oodle for buffer size
        let mut out = vec![0u8; CHUNK_SIZE];

        let compressor = OodleLZ_Compressor::Kraken;
        let level = OodleLZ_CompressionLevel::Optimal2;

        let ret = unsafe {
            let compress: Symbol<OodleLZ_Compress> = self.lib.get(b"OodleLZ_Compress")?;

            compress(
                compressor,
                raw.as_ptr() as *const _,
                raw.len(),
                out.as_mut_ptr() as *mut _,
                level,
                ptr::null_mut(),
                0,
                ptr::null_mut(),
                ptr::null_mut(),
                0,
            )
        };

        if ret == 0 {
            eyre::bail!("Failed to compress chunk.");
        }

        Ok(out)
    }
}
